import django.contrib.postgres.fields
import django.db.models.deletion
import taggit.managers
import utilities.json
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('core', '0011_move_objectchange'),
        ('extras', '0118_notifications'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ObjectChange',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('core.objectchange',),
        ),
        migrations.CreateModel(
            name='Branch',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('comments', models.TextField(blank=True)),
                ('name', models.CharField(max_length=100, unique=True)),
                ('schema_id', models.CharField(editable=False, max_length=8)),
                ('status', models.CharField(default='new', editable=False, max_length=50)),
                ('last_sync', models.DateTimeField(blank=True, editable=False, null=True)),
                ('merged_time', models.DateTimeField(blank=True, null=True)),
                ('merged_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('owner', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='branches', to=settings.AUTH_USER_MODEL)),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'branch',
                'verbose_name_plural': 'branches',
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='AppliedChange',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('change', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='application', to='core.objectchange')),
                ('branch', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='applied_changes', to='netbox_branching.branch')),
            ],
            options={
                'verbose_name': 'applied change',
                'verbose_name_plural': 'applied changes',
                'ordering': ('branch', 'change'),
            },
        ),
        migrations.CreateModel(
            name='BranchEvent',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('time', models.DateTimeField(auto_now_add=True)),
                ('type', models.CharField(editable=False, max_length=50)),
                ('branch', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='events', to='netbox_branching.branch')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='branch_events', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'branch event',
                'verbose_name_plural': 'branch events',
                'ordering': ('-time',),
            },
        ),
        migrations.CreateModel(
            name='ChangeDiff',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('last_updated', models.DateTimeField(auto_now_add=True)),
                ('object_id', models.PositiveBigIntegerField()),
                ('object_repr', models.CharField(editable=False, max_length=200)),
                ('action', models.CharField(max_length=50)),
                ('original', models.JSONField(blank=True, null=True)),
                ('modified', models.JSONField(blank=True, null=True)),
                ('current', models.JSONField(blank=True, null=True)),
                ('conflicts', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=100), blank=True, editable=False, null=True, size=None)),
                ('branch', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='netbox_branching.branch')),
                ('object_type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='contenttypes.contenttype')),
            ],
            options={
                'verbose_name': 'change diff',
                'verbose_name_plural': 'change diffs',
                'ordering': ('-last_updated',),
                'indexes': [models.Index(fields=['object_type', 'object_id'], name='netbox_bran_object__462279_idx')],
            },
        ),
    ]



from netbox.api.routers import NetBoxRouter
from . import views

router = NetBoxRouter()
router.APIRootView = views.RootView
router.register('branches', views.BranchViewSet)
router.register('branch-events', views.BranchEventViewSet)
router.register('changes', views.ChangeDiffViewSet)

urlpatterns = router.urls



from .contextvars import active_branch


__all__ = (
    'BranchAwareRouter',
)


class BranchAwareRouter:
    """
    A Django database router that returns the appropriate connection/schema for
    the active branch (if any).
    """
    def db_for_read(self, model, **hints):
        if branch := active_branch.get():
            return f'schema_{branch.schema_name}'
        return None

    def db_for_write(self, model, **hints):
        if branch := active_branch.get():
            return f'schema_{branch.schema_name}'
        return None

    def allow_relation(self, obj1, obj2, **hints):
        # Permit relations from the branch schema to the main (public) schema
        return True



import django_tables2 as tables
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from core.models import ObjectChange
from netbox.tables import NetBoxTable, columns
from netbox_branching.models import Branch, ChangeDiff
from utilities.templatetags.builtins.filters import placeholder
from .columns import ConflictsColumn, DiffColumn

__all__ = (
    'ChangeDiffTable',
    'BranchTable',
    'ChangesTable',
)


OBJECTCHANGE_FULL_NAME = """
{% load helpers %}
{{ value.get_full_name|placeholder }}
"""

OBJECTCHANGE_OBJECT = """
{% if value and value.get_absolute_url %}
    <a href="{{ value.get_absolute_url }}">{{ record.object_repr }}</a>
{% else %}
    {{ record.object_repr }}
{% endif %}
"""

BEFORE_DIFF = """
{% if record.action == 'create' %}
    {{ ''|placeholder }}
{% else %}
    <pre class="p-0">{% for k, v in record.diff.pre.items %}{{ k }}: {{ v }}
{% endfor %}</pre>
{% endif %}
"""

AFTER_DIFF = """
{% if record.action == 'delete' %}
    {{ ''|placeholder }}
{% else %}
    <pre class="p-0">{% for k, v in record.diff.post.items %}{{ k }}: {{ v }}
{% endfor %}</pre>
{% endif %}
"""

OBJECTCHANGE_REQUEST_ID = """
<a href="?request_id={{ value }}">{{ value }}</a>
"""


class BranchTable(NetBoxTable):
    name = tables.Column(
        verbose_name=_('Name'),
        linkify=True
    )
    status = columns.ChoiceFieldColumn(
        verbose_name=_('Status'),
    )
    conflicts = ConflictsColumn(
        verbose_name=_('Conflicts')
    )
    schema_id = tables.TemplateColumn(
        template_code='<span class="font-monospace">{{ value }}</code>'
    )

    class Meta(NetBoxTable.Meta):
        model = Branch
        fields = (
            'pk', 'id', 'name', 'is_active', 'status', 'conflicts', 'schema_id', 'description', 'owner', 'tags',
            'created', 'last_updated',
        )
        default_columns = (
            'pk', 'name', 'is_active', 'status', 'owner', 'conflicts', 'schema_id', 'description',
        )

    def render_is_active(self, value):
        if value:
            return mark_safe('<span class="text-success"><i class="mdi mdi-check-bold"></i></span>')
        return placeholder('')


class ChangeDiffTable(NetBoxTable):
    name = tables.Column(
        verbose_name=_('Name'),
        linkify=True
    )
    branch = tables.Column(
        verbose_name=_('Branch'),
        linkify=True
    )
    object = tables.TemplateColumn(
        template_code=OBJECTCHANGE_OBJECT,
        verbose_name=_('Object'),
        orderable=False
    )
    action = columns.ChoiceFieldColumn(
        verbose_name=_('Action'),
    )
    conflicts = ConflictsColumn(
        verbose_name=_('Conflicts')
    )
    original_diff = DiffColumn(
        show_conflicts=False,
        orderable=False,
        verbose_name=_('Main (original)')
    )
    modified_diff = DiffColumn(
        orderable=False,
        verbose_name=_('Branch (current)')
    )
    current_diff = DiffColumn(
        orderable=False,
        verbose_name=_('Main (current)')
    )
    actions = columns.ActionsColumn(
        actions=()
    )

    class Meta(NetBoxTable.Meta):
        model = ChangeDiff
        fields = (
            'branch', 'object_type', 'object', 'action', 'conflicts', 'original_diff', 'modified_diff', 'current_diff',
            'last_updated', 'actions',
        )
        default_columns = ('branch', 'object', 'action', 'conflicts', 'original_diff', 'modified_diff', 'current_diff')


class ChangesTable(NetBoxTable):
    time = columns.DateTimeColumn(
        verbose_name=_('Time'),
        timespec='minutes',
        linkify=True
    )
    action = columns.ChoiceFieldColumn(
        verbose_name=_('Action'),
    )
    model = tables.Column()
    changed_object_type = columns.ContentTypeColumn(
        verbose_name=_('Type')
    )
    object_repr = tables.TemplateColumn(
        accessor=tables.A('changed_object'),
        template_code=OBJECTCHANGE_OBJECT,
        verbose_name=_('Object'),
        orderable=False
    )
    before = tables.TemplateColumn(
        accessor=tables.A('prechange_data_clean'),
        template_code=BEFORE_DIFF,
        verbose_name=_('Before'),
        orderable=False
    )
    after = tables.TemplateColumn(
        accessor=tables.A('postchange_data_clean'),
        template_code=AFTER_DIFF,
        verbose_name=_('After'),
        orderable=False
    )
    request_id = tables.TemplateColumn(
        template_code=OBJECTCHANGE_REQUEST_ID,
        verbose_name=_('Request ID')
    )
    actions = columns.ActionsColumn(
        actions=()
    )

    class Meta(NetBoxTable.Meta):
        model = ObjectChange
        fields = (
            'pk', 'time', 'action', 'model', 'changed_object_type', 'object_repr', 'request_id', 'before', 'after',
        )



import datetime
import logging
from contextlib import contextmanager
from dataclasses import dataclass

from django.db.models import ForeignKey, ManyToManyField

from .contextvars import active_branch

__all__ = (
    'ChangeSummary',
    'DynamicSchemaDict',
    'ListHandler',
    'activate_branch',
    'deactivate_branch',
    'get_branchable_object_types',
    'get_tables_to_replicate',
    'record_applied_change',
    'update_object',
)


class DynamicSchemaDict(dict):
    """
    Behaves like a normal dictionary, except for keys beginning with "schema_". Any lookup for
    "schema_*" will return the default configuration extended to include the search_path option.
    """
    def __getitem__(self, item):
        if type(item) is str and item.startswith('schema_'):
            if schema := item.removeprefix('schema_'):
                default_config = super().__getitem__('default')
                return {
                    **default_config,
                    'OPTIONS': {
                        'options': f'-c search_path={schema},public'
                    }
                }
        return super().__getitem__(item)

    def __contains__(self, item):
        if type(item) is str and item.startswith('schema_'):
            return True
        return super().__contains__(item)


@contextmanager
def activate_branch(branch):
    """
    A context manager for activating a Branch.
    """
    token = active_branch.set(branch)

    yield

    active_branch.reset(token)


@contextmanager
def deactivate_branch():
    """
    A context manager for temporarily deactivating the active Branch (if any). This is a
    convenience function for `activate_branch(None)`.
    """
    token = active_branch.set(None)

    yield

    active_branch.reset(token)


def get_branchable_object_types():
    """
    Return all object types which are branch-aware; i.e. those which support change logging.
    """
    from core.models import ObjectType
    return ObjectType.objects.with_feature('change_logging').exclude(
        app_label__in=['netbox_branching', 'netbox_changes']
    )


def get_tables_to_replicate():
    """
    Returned an ordered list of database tables to replicate when provisioning a new schema.
    """
    tables = set()

    branch_aware_models = [
        ot.model_class() for ot in get_branchable_object_types()
    ]
    for model in branch_aware_models:

        # Capture the model's table
        tables.add(model._meta.db_table)

        # Capture any M2M fields which reference other replicated models
        for m2m_field in model._meta.local_many_to_many:
            if m2m_field.related_model in branch_aware_models:
                if hasattr(m2m_field, 'through'):
                    # Field is actually a manager
                    m2m_table = m2m_field.through._meta.db_table
                else:
                    m2m_table = m2m_field._get_m2m_db_table(model._meta)
                tables.add(m2m_table)

    return sorted(tables)


class ListHandler(logging.Handler):
    """
    A logging handler which appends log messages to list passed on initialization.
    """
    def __init__(self, *args, queue, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue = queue

    def emit(self, record):
        self.queue.append(self.format(record))


@dataclass
class ChangeSummary:
    """
    A record indicating the number of changes which were made between a start and end time.
    """
    start: datetime.datetime
    end: datetime.datetime
    count: int


def update_object(instance, data, using):
    """
    Set an attribute on an object depending on the type of model field.
    """
    # Avoid AppRegistryNotReady exception
    from taggit.managers import TaggableManager
    instance.snapshot()
    m2m_assignments = {}

    for attr, value in data.items():
        model_field = instance._meta.get_field(attr)
        field_cls = model_field.__class__

        if issubclass(field_cls, ForeignKey):
            # Direct value assignment for ForeignKeys must be done by the field's concrete name
            setattr(instance, f'{attr}_id', value)
        elif issubclass(field_cls, (ManyToManyField, TaggableManager)):
            # Use M2M manager for ManyToMany assignments
            m2m_manager = getattr(instance, attr)
            m2m_assignments[m2m_manager] = value
        else:
            setattr(instance, attr, value)

    instance.full_clean()
    instance.save(using=using)
    for m2m_manager, value in m2m_assignments.items():
        m2m_manager.set(value)


def record_applied_change(instance, branch, **kwargs):
    """
    Create a new AppliedChange instance mapping an applied ObjectChange to its Branch.
    """
    from .models import AppliedChange

    AppliedChange.objects.update_or_create(change=instance, defaults={'branch': branch})



from django.contrib.contenttypes.models import ContentType
from netbox.plugins import PluginTemplateExtension

from .choices import BranchStatusChoices
from .contextvars import active_branch
from .models import Branch, ChangeDiff

__all__ = (
    'BranchNotification',
    'BranchSelector',
    'template_extensions',
)


class BranchSelector(PluginTemplateExtension):

    def navbar(self):
        return self.render('netbox_branching/inc/branch_selector.html', extra_context={
            'active_branch': active_branch.get(),
            'branches': Branch.objects.exclude(status=BranchStatusChoices.MERGED),
        })


class BranchNotification(PluginTemplateExtension):

    def alerts(self):
        if not (instance := self.context['object']):
            return ''
        ct = ContentType.objects.get_for_model(instance)
        relevant_changes = ChangeDiff.objects.filter(
            object_type=ct,
            object_id=instance.pk
        ).exclude(
            branch__status=BranchStatusChoices.MERGED
        )
        branches = [
            diff.branch for diff in relevant_changes.only('branch')
        ]
        return self.render('netbox_branching/inc/modified_notice.html', extra_context={
            'branches': branches,
        })


template_extensions = [BranchSelector, BranchNotification]



from netbox.plugins import PluginConfig


class AppConfig(PluginConfig):
    name = 'netbox_branching'
    verbose_name = 'NetBox Branching'
    description = 'A git-like branching implementation for NetBox'
    version = '0.3.1'
    base_url = 'branching'
    min_version = '4.1'
    middleware = [
        'netbox_branching.middleware.BranchMiddleware'
    ]
    default_settings = {
        # This string is prefixed to the name of each new branch schema during provisioning
        'schema_prefix': 'branch_',
    }

    def ready(self):
        super().ready()
        from . import events, search, signal_receivers


config = AppConfig



from django.dispatch import Signal

from .events import *

__all__ = (
    'branch_deprovisioned',
    'branch_merged',
    'branch_provisioned',
    'branch_reverted',
    'branch_synced',
)


branch_provisioned = Signal()
branch_deprovisioned = Signal()
branch_synced = Signal()
branch_merged = Signal()
branch_reverted = Signal()

branch_signals = {
    branch_provisioned: BRANCH_PROVISIONED,
    branch_deprovisioned: BRANCH_DEPROVISIONED,
    branch_synced: BRANCH_SYNCED,
    branch_merged: BRANCH_MERGED,
    branch_reverted: BRANCH_REVERTED,
}



from django.contrib import messages
from django.contrib.contenttypes.models import ContentType
from django.db.models import Count, Q
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as _

from core.choices import ObjectChangeActionChoices
from core.filtersets import ObjectChangeFilterSet
from core.models import ObjectChange
from netbox.views import generic
from utilities.views import ViewTab, register_model_view
from . import filtersets, forms, tables
from .choices import BranchStatusChoices
from .jobs import MergeBranchJob, RevertBranchJob, SyncBranchJob
from .models import ChangeDiff, Branch


#
# Branches
#

class BranchListView(generic.ObjectListView):
    queryset = Branch.objects.annotate(
        # Annotate the number of associated ChangeDiffs with conflicts
        conflicts=Count('changediff', filter=Q(changediff__conflicts__isnull=False))
    ).order_by('name')
    filterset = filtersets.BranchFilterSet
    filterset_form = forms.BranchFilterForm
    table = tables.BranchTable


@register_model_view(Branch)
class BranchView(generic.ObjectView):
    queryset = Branch.objects.all()

    def get_extra_context(self, request, instance):
        qs = instance.get_changes().values_list('changed_object_type').annotate(count=Count('pk'))
        if instance.ready or instance.merged:
            stats = {
                'created': {
                    ContentType.objects.get(pk=ct): count
                    for ct, count in qs.filter(action=ObjectChangeActionChoices.ACTION_CREATE)
                },
                'updated': {
                    ContentType.objects.get(pk=ct): count
                    for ct, count in qs.filter(action=ObjectChangeActionChoices.ACTION_UPDATE)
                },
                'deleted': {
                    ContentType.objects.get(pk=ct): count
                    for ct, count in qs.filter(action=ObjectChangeActionChoices.ACTION_DELETE)
                },
            }
        else:
            stats = {}

        latest_change = instance.get_changes().order_by('time').last()
        last_job = instance.jobs.order_by('created').last()

        return {
            'stats': stats,
            'latest_change': latest_change,
            'last_job': last_job,
            'conflicts_count': ChangeDiff.objects.filter(branch=instance, conflicts__isnull=False).count(),
        }


@register_model_view(Branch, 'edit')
class BranchEditView(generic.ObjectEditView):
    queryset = Branch.objects.all()
    form = forms.BranchForm

    def alter_object(self, obj, request, url_args, url_kwargs):
        if not obj.pk:
            obj.owner = request.user
        return obj


@register_model_view(Branch, 'delete')
class BranchDeleteView(generic.ObjectDeleteView):
    queryset = Branch.objects.all()
    default_return_url = 'plugins:netbox_branching:branch_list'


def _get_diff_count(obj):
    return ChangeDiff.objects.filter(branch=obj).count()


@register_model_view(Branch, 'diff')
class BranchDiffView(generic.ObjectChildrenView):
    queryset = Branch.objects.all()
    child_model = ChangeDiff
    filterset = filtersets.ChangeDiffFilterSet
    table = tables.ChangeDiffTable
    actions = {}
    tab = ViewTab(
        label=_('Diff'),
        badge=_get_diff_count,
        permission='netbox_branching.view_branch'
    )

    def get_children(self, request, parent):
        return ChangeDiff.objects.filter(branch=parent)


@register_model_view(Branch, 'changes-behind')
class BranchChangesBehindView(generic.ObjectChildrenView):
    queryset = Branch.objects.all()
    child_model = ObjectChange
    filterset = ObjectChangeFilterSet
    table = tables.ChangesTable
    actions = {}
    tab = ViewTab(
        label=_('Changes Behind'),
        badge=lambda obj: obj.get_unsynced_changes().count(),
        permission='netbox_branching.view_branch'
    )

    def get_children(self, request, parent):
        return parent.get_unsynced_changes().order_by('time')


@register_model_view(Branch, 'changes-ahead')
class BranchChangesAheadView(generic.ObjectChildrenView):
    queryset = Branch.objects.all()
    child_model = ObjectChange
    filterset = ObjectChangeFilterSet
    table = tables.ChangesTable
    actions = {}
    tab = ViewTab(
        label=_('Changes Ahead'),
        badge=lambda obj: obj.get_unmerged_changes().count(),
        permission='netbox_branching.view_branch'
    )

    def get_children(self, request, parent):
        return parent.get_unmerged_changes().order_by('time')


def _get_change_count(obj):
    return obj.get_unmerged_changes().count()


@register_model_view(Branch, 'changes-merged')
class BranchChangesAheadView(generic.ObjectChildrenView):
    queryset = Branch.objects.all()
    child_model = ObjectChange
    filterset = ObjectChangeFilterSet
    table = tables.ChangesTable
    actions = {}
    tab = ViewTab(
        label=_('Changes Merged'),
        badge=lambda obj: obj.get_merged_changes().count(),
        permission='netbox_branching.view_branch',
        hide_if_empty=True
    )

    def get_children(self, request, parent):
        return parent.get_merged_changes().order_by('time')


class BaseBranchActionView(generic.ObjectView):
    """
    Base view for syncing or merging a Branch.
    """
    queryset = Branch.objects.all()
    form = forms.BranchActionForm
    template_name = 'netbox_branching/branch_action.html'
    action = None
    valid_states = (
        BranchStatusChoices.READY,
    )

    def get_required_permission(self):
        return f'netbox_branching.{self.action}_branch'

    @staticmethod
    def _get_conflicts_table(branch):
        conflicts = ChangeDiff.objects.filter(branch=branch, conflicts__isnull=False)
        conflicts_table = tables.ChangeDiffTable(conflicts)
        conflicts_table.columns.show('pk')

        return conflicts_table

    def do_action(self, branch, request, form):
        raise NotImplementedError(f"{self.__class__} must implement action() method.")

    def get(self, request, **kwargs):
        branch = self.get_object(**kwargs)
        form = forms.BranchActionForm(branch)

        return render(request, self.template_name, {
            'branch': branch,
            'action': _(f'{self.action.title()} Branch'),
            'form': form,
            'conflicts_table': self._get_conflicts_table(branch),
        })

    def post(self, request, **kwargs):
        branch = self.get_object(**kwargs)
        form = forms.BranchActionForm(branch, request.POST)

        if branch.status not in self.valid_states:
            messages.error(request, _(
                "The branch must be in one of the following states to perform this action: {valid_states}"
            ).format(valid_states=', '.join(self.valid_states)))
        elif form.is_valid():
            return self.do_action(branch, request, form)

        return render(request, self.template_name, {
            'branch': branch,
            'action': _(f'{self.action.title()} Branch'),
            'form': form,
            'conflicts_table': self._get_conflicts_table(branch),
        })


@register_model_view(Branch, 'sync')
class BranchSyncView(BaseBranchActionView):
    action = 'sync'

    def do_action(self, branch, request, form):
        # Enqueue a background job to sync the Branch
        SyncBranchJob.enqueue(
            instance=branch,
            user=request.user,
            commit=form.cleaned_data['commit']
        )
        messages.success(request, f"Syncing of branch {branch} in progress")

        return redirect(branch.get_absolute_url())


@register_model_view(Branch, 'merge')
class BranchMergeView(BaseBranchActionView):
    action = 'merge'

    def do_action(self, branch, request, form):
        # Enqueue a background job to merge the Branch
        MergeBranchJob.enqueue(
            instance=branch,
            user=request.user,
            commit=form.cleaned_data['commit']
        )
        messages.success(request, f"Merging of branch {branch} in progress")

        return redirect(branch.get_absolute_url())


@register_model_view(Branch, 'revert')
class BranchRevertView(BaseBranchActionView):
    action = 'revert'
    valid_states = (
        BranchStatusChoices.MERGED,
    )

    def do_action(self, branch, request, form):
        # Enqueue a background job to revert the Branch
        RevertBranchJob.enqueue(
            instance=branch,
            user=request.user,
            commit=form.cleaned_data['commit']
        )
        messages.success(request, f"Reverting branch {branch}")

        return redirect(branch.get_absolute_url())


class BranchBulkImportView(generic.BulkImportView):
    queryset = Branch.objects.all()
    model_form = forms.BranchImportForm


class BranchBulkEditView(generic.BulkEditView):
    queryset = Branch.objects.all()
    filterset = filtersets.BranchFilterSet
    table = tables.BranchTable
    form = forms.BranchBulkEditForm


class BranchBulkDeleteView(generic.BulkDeleteView):
    queryset = Branch.objects.all()
    filterset = filtersets.BranchFilterSet
    table = tables.BranchTable


#
# Change diffs
#

class ChangeDiffListView(generic.ObjectListView):
    queryset = ChangeDiff.objects.all()
    filterset = filtersets.ChangeDiffFilterSet
    filterset_form = forms.ChangeDiffFilterForm
    table = tables.ChangeDiffTable



import logging
import random
import string
from functools import cached_property, partial

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import DEFAULT_DB_ALIAS, connection, models, transaction
from django.db.models.signals import post_save
from django.db.utils import ProgrammingError
from django.test import RequestFactory
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from core.models import ObjectChange as ObjectChange_
from netbox.context import current_request
from netbox.context_managers import event_tracking
from netbox.models import PrimaryModel
from netbox.models.features import JobsMixin
from netbox.plugins import get_plugin_config
from netbox_branching.choices import BranchEventTypeChoices, BranchStatusChoices
from netbox_branching.contextvars import active_branch
from netbox_branching.signals import *
from netbox_branching.utilities import (
    ChangeSummary, activate_branch, get_branchable_object_types, get_tables_to_replicate, record_applied_change,
)
from utilities.exceptions import AbortRequest, AbortTransaction
from .changes import ObjectChange

__all__ = (
    'Branch',
    'BranchEvent',
)


class Branch(JobsMixin, PrimaryModel):
    name = models.CharField(
        verbose_name=_('name'),
        max_length=100,
        unique=True
    )
    owner = models.ForeignKey(
        to=get_user_model(),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='branches'
    )
    schema_id = models.CharField(
        max_length=8,
        unique=True,
        verbose_name=_('schema ID'),
        editable=False
    )
    status = models.CharField(
        verbose_name=_('status'),
        max_length=50,
        choices=BranchStatusChoices,
        default=BranchStatusChoices.NEW,
        editable=False
    )
    last_sync = models.DateTimeField(
        blank=True,
        null=True,
        editable=False
    )
    merged_time = models.DateTimeField(
        verbose_name=_('merged time'),
        blank=True,
        null=True
    )
    merged_by = models.ForeignKey(
        to=get_user_model(),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='+'
    )

    class Meta:
        ordering = ('name',)
        verbose_name = _('branch')
        verbose_name_plural = _('branches')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Generate a random schema ID if this is a new Branch
        if self.pk is None:
            self.schema_id = self._generate_schema_id()

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('plugins:netbox_branching:branch', args=[self.pk])

    def get_status_color(self):
        return BranchStatusChoices.colors.get(self.status)

    @cached_property
    def is_active(self):
        return self == active_branch.get()

    @property
    def ready(self):
        return self.status == BranchStatusChoices.READY

    @property
    def merged(self):
        return self.status == BranchStatusChoices.MERGED

    @cached_property
    def schema_name(self):
        schema_prefix = get_plugin_config('netbox_branching', 'schema_prefix')
        return f'{schema_prefix}{self.schema_id}'

    @cached_property
    def connection_name(self):
        return f'schema_{self.schema_name}'

    @cached_property
    def synced_time(self):
        return self.last_sync or self.created

    def save(self, provision=True, *args, **kwargs):
        """
        Args:
            provision: If True, automatically enqueue a background Job to provision the Branch. (Set this
                       to False if you will call provision() on the instance manually.)
        """
        from netbox_branching.jobs import ProvisionBranchJob

        _provision = provision and self.pk is None

        if active_branch.get():
            raise AbortRequest(_("Cannot create or modify a branch while a branch is active."))

        super().save(*args, **kwargs)

        if _provision:
            # Enqueue a background job to provision the Branch
            request = current_request.get()
            ProvisionBranchJob.enqueue(
                instance=self,
                user=request.user if request else None
            )

    def delete(self, *args, **kwargs):
        if active_branch.get():
            raise AbortRequest(_("Cannot delete a branch while a branch is active."))

        # Deprovision the schema
        self.deprovision()

        return super().delete(*args, **kwargs)

    @staticmethod
    def _generate_schema_id(length=8):
        """
        Generate a random alphanumeric schema identifier of the specified length.
        """
        chars = [*string.ascii_lowercase, *string.digits]
        return ''.join(random.choices(chars, k=length))

    def get_changes(self):
        """
        Return a queryset of all ObjectChange records created within the Branch.
        """
        if self.status == BranchStatusChoices.NEW:
            return ObjectChange.objects.none()
        return ObjectChange.objects.using(self.connection_name)

    def get_unsynced_changes(self):
        """
        Return a queryset of all ObjectChange records created in main since the Branch was last synced or created.
        """
        return ObjectChange.objects.using(DEFAULT_DB_ALIAS).exclude(
            application__branch=self
        ).filter(
            changed_object_type__in=get_branchable_object_types(),
            time__gt=self.synced_time
        )

    def get_unmerged_changes(self):
        """
        Return a queryset of all unmerged ObjectChange records within the Branch schema.
        """
        if self.status == BranchStatusChoices.MERGED:
            return ObjectChange.objects.none()
        return ObjectChange.objects.using(self.connection_name)

    def get_merged_changes(self):
        """
        Return a queryset of all merged ObjectChange records for the Branch.
        """
        if self.status != BranchStatusChoices.MERGED:
            return ObjectChange.objects.none()
        return ObjectChange.objects.using(DEFAULT_DB_ALIAS).filter(
            application__branch=self
        )

    def get_event_history(self):
        history = []
        last_time = timezone.now()
        for event in self.events.all():
            if change_count := self.get_changes().filter(time__gte=event.time, time__lt=last_time).count():
                summary = ChangeSummary(
                    start=event.time,
                    end=last_time,
                    count=change_count
                )
                history.append(summary)
            history.append(event)
            last_time = event.time
        return history

    def sync(self, user, commit=True):
        """
        Apply changes from the main schema onto the Branch's schema.
        """
        logger = logging.getLogger('netbox_branching.branch.sync')
        logger.info(f'Syncing branch {self} ({self.schema_name})')

        if not self.ready:
            raise Exception(f"Branch {self} is not ready to sync")

        # Update Branch status
        Branch.objects.filter(pk=self.pk).update(status=BranchStatusChoices.SYNCING)

        try:
            with activate_branch(self):
                with transaction.atomic():
                    # Apply each change from the main schema
                    for change in self.get_unsynced_changes().order_by('time'):
                        logger.debug(f'Applying change: {change}')
                        change.apply(using=self.connection_name)
                    if not commit:
                        raise AbortTransaction()

        except Exception as e:
            logger.error(e)
            # Restore original branch status
            Branch.objects.filter(pk=self.pk).update(status=BranchStatusChoices.READY)
            raise e

        # Record the branch's last_synced time & update its status
        self.last_sync = timezone.now()
        self.status = BranchStatusChoices.READY
        self.save()
        BranchEvent.objects.create(branch=self, user=user, type=BranchEventTypeChoices.SYNCED)

        # Emit branch_synced signal
        branch_synced.send(sender=self.__class__, branch=self, user=user)

        logger.info('Syncing completed')

    sync.alters_data = True

    def merge(self, user, commit=True):
        """
        Apply all changes in the Branch to the main schema by replaying them in
        chronological order.
        """
        logger = logging.getLogger('netbox_branching.branch.merge')
        logger.info(f'Merging branch {self} ({self.schema_name})')

        if not self.ready:
            raise Exception(f"Branch {self} is not ready to merge")

        # Retrieve staged changes before we update the Branch's status
        changes = self.get_unmerged_changes().order_by('time')

        # Update Branch status
        Branch.objects.filter(pk=self.pk).update(status=BranchStatusChoices.MERGING)

        # Create a dummy request for the event_tracking() context manager
        request = RequestFactory().get(reverse('home'))

        # Prep & connect the signal receiver for recording AppliedChanges
        handler = partial(record_applied_change, branch=self)
        post_save.connect(handler, sender=ObjectChange_, weak=False)

        try:
            with transaction.atomic():
                # Apply each change from the Branch
                for change in changes:
                    with event_tracking(request):
                        logger.debug(f'Applying change: {change}')
                        request.id = change.request_id
                        request.user = change.user
                        change.apply(using=DEFAULT_DB_ALIAS)
                if not commit:
                    raise AbortTransaction()

        except Exception as e:
            logger.error(e)
            # Disconnect signal receiver & restore original branch status
            post_save.disconnect(handler, sender=ObjectChange_)
            Branch.objects.filter(pk=self.pk).update(status=BranchStatusChoices.READY)
            raise e

        # Update the Branch's status to "merged"
        self.status = BranchStatusChoices.MERGED
        self.merged_time = timezone.now()
        self.merged_by = user
        self.save()
        BranchEvent.objects.create(branch=self, user=user, type=BranchEventTypeChoices.MERGED)

        # Emit branch_merged signal
        branch_merged.send(sender=self.__class__, branch=self, user=user)

        logger.info('Merging completed')

        # Disconnect the signal receiver
        post_save.disconnect(handler, sender=ObjectChange_)

    merge.alters_data = True

    def revert(self, user, commit=True):
        """
        Undo all changes associated with a previously merged Branch in the main schema by replaying them in
        reverse order and calling undo() on each.
        """
        logger = logging.getLogger('netbox_branching.branch.revert')
        logger.info(f'Reverting branch {self} ({self.schema_name})')

        if not self.merged:
            raise Exception(f"Only merged branches can be reverted.")

        # Retrieve applied changes before we update the Branch's status
        changes = self.get_changes().order_by('-time')

        # Update Branch status
        Branch.objects.filter(pk=self.pk).update(status=BranchStatusChoices.REVERTING)

        # Create a dummy request for the event_tracking() context manager
        request = RequestFactory().get(reverse('home'))

        # Prep & connect the signal receiver for recording AppliedChanges
        handler = partial(record_applied_change, branch=self)
        post_save.connect(handler, sender=ObjectChange_, weak=False)

        try:
            with transaction.atomic():
                # Undo each change from the Branch
                for change in changes:
                    with event_tracking(request):
                        logger.debug(f'Undoing change: {change}')
                        request.id = change.request_id
                        request.user = change.user
                        change.undo()
                if not commit:
                    raise AbortTransaction()

        except Exception as e:
            logger.error(e)
            # Disconnect signal receiver & restore original branch status
            post_save.disconnect(handler, sender=ObjectChange_)
            Branch.objects.filter(pk=self.pk).update(status=BranchStatusChoices.MERGED)
            raise e

        # Update the Branch's status to "ready"
        self.status = BranchStatusChoices.READY
        self.merged_time = None
        self.merged_by = None
        self.save()
        BranchEvent.objects.create(branch=self, user=user, type=BranchEventTypeChoices.REVERTED)

        # Emit branch_reverted signal
        branch_reverted.send(sender=self.__class__, branch=self, user=user)

        logger.info('Reversion completed')

        # Disconnect the signal receiver
        post_save.disconnect(handler, sender=ObjectChange_)

    merge.alters_data = True

    def provision(self, user):
        """
        Create the schema & replicate main tables.
        """
        logger = logging.getLogger('netbox_branching.branch.provision')
        logger.info(f'Provisioning branch {self} ({self.schema_name})')

        # Update Branch status
        Branch.objects.filter(pk=self.pk).update(status=BranchStatusChoices.PROVISIONING)

        with connection.cursor() as cursor:
            try:
                schema = self.schema_name

                # Start a transaction
                cursor.execute("BEGIN")
                cursor.execute("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE")

                # Create the new schema
                logger.debug(f'Creating schema {schema}')
                try:
                    cursor.execute(f"CREATE SCHEMA {schema}")
                except ProgrammingError as e:
                    if str(e).startswith('permission denied '):
                        logger.critical(
                            f"Provisioning failed due to insufficient database permissions. Ensure that the NetBox "
                            f"role ({settings.DATABASE['USER']}) has permission to create new schemas on this "
                            f"database ({settings.DATABASE['NAME']}). (Use the PostgreSQL command 'GRANT CREATE ON "
                            f"DATABASE $database TO $role;' to grant the required permission.)"
                        )
                    raise e

                # Create an empty copy of the global change log. Share the ID sequence from the main table to avoid
                # reusing change record IDs.
                table = ObjectChange_._meta.db_table
                logger.debug(f'Creating table {schema}.{table}')
                cursor.execute(
                    f"CREATE TABLE {schema}.{table} ( LIKE public.{table} INCLUDING INDEXES )"
                )
                # Set the default value for the ID column to the sequence associated with the source table
                cursor.execute(
                    f"ALTER TABLE {schema}.{table} "
                    f"ALTER COLUMN id SET DEFAULT nextval('public.{table}_id_seq')"
                )

                # Replicate relevant tables from the main schema
                for table in get_tables_to_replicate():
                    logger.debug(f'Creating table {schema}.{table}')
                    # Create the table in the new schema
                    cursor.execute(
                        f"CREATE TABLE {schema}.{table} ( LIKE public.{table} INCLUDING INDEXES )"
                    )
                    # Copy data from the source table
                    cursor.execute(
                        f"INSERT INTO {schema}.{table} SELECT * FROM public.{table}"
                    )
                    # Get the name of the sequence used for object ID allocations
                    cursor.execute(
                        f"SELECT pg_get_serial_sequence('{table}', 'id');"
                    )
                    sequence_name = cursor.fetchone()[0]
                    # Set the default value for the ID column to the sequence associated with the source table
                    cursor.execute(
                        f"ALTER TABLE {schema}.{table} ALTER COLUMN id SET DEFAULT nextval('{sequence_name}')"
                    )

                # Commit the transaction
                cursor.execute("COMMIT")

            except Exception as e:
                # Abort the transaction
                cursor.execute("ROLLBACK")

                # Mark the Branch as failed
                logger.error(e)
                Branch.objects.filter(pk=self.pk).update(status=BranchStatusChoices.FAILED)

                raise e

        # Emit branch_provisioned signal
        branch_provisioned.send(sender=self.__class__, branch=self, user=user)

        logger.info('Provisioning completed')

        Branch.objects.filter(pk=self.pk).update(status=BranchStatusChoices.READY)
        BranchEvent.objects.create(branch=self, user=user, type=BranchEventTypeChoices.PROVISIONED)

    provision.alters_data = True

    def deprovision(self):
        """
        Delete the Branch's schema and all its tables from the database.
        """
        logger = logging.getLogger('netbox_branching.branch.provision')
        logger.info(f'Deprovisioning branch {self} ({self.schema_name})')

        with connection.cursor() as cursor:
            # Delete the schema and all its tables
            cursor.execute(
                f"DROP SCHEMA IF EXISTS {self.schema_name} CASCADE"
            )

        # Emit branch_deprovisioned signal
        branch_deprovisioned.send(sender=self.__class__, branch=self)

        logger.info('Deprovisioning completed')

    deprovision.alters_data = True


class BranchEvent(models.Model):
    time = models.DateTimeField(
        auto_now_add=True,
        editable=False
    )
    branch = models.ForeignKey(
        to='netbox_branching.branch',
        on_delete=models.CASCADE,
        related_name='events'
    )
    user = models.ForeignKey(
        to=get_user_model(),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name='branch_events'
    )
    type = models.CharField(
        verbose_name=_('type'),
        max_length=50,
        choices=BranchEventTypeChoices,
        editable=False
    )

    class Meta:
        ordering = ('-time',)
        verbose_name = _('branch event')
        verbose_name_plural = _('branch events')

    def get_type_color(self):
        return BranchEventTypeChoices.colors.get(self.type)



from netbox_branching.models import Branch

from netbox.forms import NetBoxModelImportForm

__all__ = (
    'BranchImportForm',
)


class BranchImportForm(NetBoxModelImportForm):

    class Meta:
        model = Branch
        fields = (
            'name', 'description', 'comments', 'tags',
        )



import django_filters
from django.db.models import Q
from django.utils.translation import gettext as _

from core.choices import ObjectChangeActionChoices
from core.models import ObjectType
from netbox.filtersets import BaseFilterSet, NetBoxModelFilterSet
from utilities import filters
from .choices import *
from .models import *

__all__ = (
    'BranchEventFilterSet',
    'BranchFilterSet',
    'ChangeDiffFilterSet',
)


class BranchFilterSet(NetBoxModelFilterSet):
    status = django_filters.MultipleChoiceFilter(
        choices=BranchStatusChoices,
        null_value=None
    )
    last_sync = filters.MultiValueDateTimeFilter()

    class Meta:
        model = Branch
        fields = ('id', 'name', 'description')

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) |
            Q(description__icontains=value)
        )


class BranchEventFilterSet(BaseFilterSet):
    type = django_filters.MultipleChoiceFilter(
        choices=BranchEventTypeChoices,
        null_value=None
    )
    time = filters.MultiValueDateTimeFilter()

    class Meta:
        model = BranchEvent
        fields = ('id',)


class ChangeDiffFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )
    branch_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Branch.objects.all(),
        label=_('Branch (ID)'),
    )
    branch = django_filters.ModelMultipleChoiceFilter(
        field_name='branch__schema_id',
        queryset=Branch.objects.all(),
        to_field_name='schema_id',
        label=_('Branch (schema ID)'),
    )
    last_updated = filters.MultiValueDateTimeFilter()
    object_type_id = django_filters.ModelMultipleChoiceFilter(
        queryset=ObjectType.objects.all(),
        field_name='object_type'
    )
    object_type = filters.ContentTypeFilter()
    action = django_filters.MultipleChoiceFilter(
        choices=ObjectChangeActionChoices,
        null_value=None
    )
    has_conflicts = django_filters.BooleanFilter(
        method='_has_conflicts'
    )

    class Meta:
        model = ChangeDiff
        fields = ('id', 'object_type', 'object_id')

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(object_repr__icontains=value)
        )

    def _has_conflicts(self, queryset, name, value):
        if value:
            return queryset.filter(conflicts__isnull=False)
        return queryset.filter(conflicts__isnull=True)



from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseBadRequest
from django.urls import reverse

from utilities.api import is_api_request

from .choices import BranchStatusChoices
from .constants import COOKIE_NAME, BRANCH_HEADER, QUERY_PARAM
from .models import Branch
from .utilities import activate_branch

__all__ = (
    'BranchMiddleware',
)


class BranchMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):

        # Set/clear the active Branch on the request
        try:
            branch = self.get_active_branch(request)
        except ObjectDoesNotExist:
            return HttpResponseBadRequest("Invalid branch identifier")

        with activate_branch(branch):
            response = self.get_response(request)

        # Set/clear the branch cookie (for non-API requests)
        if not is_api_request(request):
            if branch:
                response.set_cookie(COOKIE_NAME, branch.schema_id)
            elif QUERY_PARAM in request.GET:
                response.delete_cookie(COOKIE_NAME)

        return response

    @staticmethod
    def get_active_branch(request):
        """
        Return the active Branch (if any).
        """
        # The active Branch is specified by HTTP header for REST API requests.
        if request.path_info.startswith(reverse('api-root')) and (schema_id := request.headers.get(BRANCH_HEADER)):
            branch = Branch.objects.get(schema_id=schema_id)
            if not branch.ready:
                return HttpResponseBadRequest(f"Branch {branch} is not ready for use (status: {branch.status})")
            return branch

        # Branch activated/deactivated by URL query parameter
        elif QUERY_PARAM in request.GET:
            if schema_id := request.GET.get(QUERY_PARAM):
                branch = Branch.objects.get(schema_id=schema_id)
                if branch.ready:
                    messages.success(request, f"Activated branch {branch}")
                    return branch
                else:
                    messages.error(request, f"Branch {branch} is not ready for use (status: {branch.status})")
                    return None
            else:
                messages.success(request, f"Deactivated branch")
                request.COOKIES.pop(COOKIE_NAME, None)  # Delete cookie if set
                return None

        # Branch set by cookie
        elif schema_id := request.COOKIES.get(COOKIE_NAME):
            return Branch.objects.filter(schema_id=schema_id, status=BranchStatusChoices.READY).first()



from django import forms
from django.utils.translation import gettext as _

from core.choices import ObjectChangeActionChoices
from core.models import ObjectType
from netbox.forms import NetBoxModelFilterSetForm
from netbox_branching.choices import BranchStatusChoices
from netbox_branching.models import ChangeDiff, Branch
from utilities.forms.fields import ContentTypeMultipleChoiceField, DynamicModelMultipleChoiceField, TagFilterField
from utilities.forms.rendering import FieldSet

__all__ = (
    'ChangeDiffFilterForm',
    'BranchFilterForm',
)


class BranchFilterForm(NetBoxModelFilterSetForm):
    model = Branch
    fieldsets = (
        FieldSet('q', 'filter_id', 'tag'),
        FieldSet('status', 'last_sync', name=_('Branch')),
    )
    status = forms.MultipleChoiceField(
        label=_('Status'),
        choices=BranchStatusChoices,
        required=False
    )
    tag = TagFilterField(model)


class ChangeDiffFilterForm(NetBoxModelFilterSetForm):
    model = ChangeDiff
    fieldsets = (
        FieldSet('filter_id',),
        FieldSet('branch_id', 'object_type_id', 'action', name=_('Change')),
    )
    branch_id = DynamicModelMultipleChoiceField(
        queryset=Branch.objects.all(),
        required=False,
        label=_('Branch')
    )
    object_type_id = ContentTypeMultipleChoiceField(
        queryset=ObjectType.objects.with_feature('change_logging'),
        required=False,
        label=_('Object Type'),
    )
    action = forms.MultipleChoiceField(
        label=_('Action'),
        choices=ObjectChangeActionChoices,
        required=False
    )



import logging

from django.db.models.signals import m2m_changed, post_save, pre_delete

from core.signals import handle_changed_object, handle_deleted_object
from netbox.jobs import JobRunner
from utilities.exceptions import AbortTransaction
from .utilities import ListHandler

__all__ = (
    'MergeBranchJob',
    'ProvisionBranchJob',
    'RevertBranchJob',
    'SyncBranchJob',
)


def get_job_log(job):
    """
    Initialize and return the job log.
    """
    job.data = {
        'log': list()
    }
    return job.data['log']


class ProvisionBranchJob(JobRunner):
    """
    Provision a Branch in the database.
    """
    class Meta:
        name = 'Provision branch'

    def run(self, *args, **kwargs):
        # Initialize logging
        logger = logging.getLogger('netbox_branching.branch.provision')
        logger.setLevel(logging.DEBUG)
        logger.addHandler(ListHandler(queue=get_job_log(self.job)))

        # Provision the Branch
        branch = self.job.object
        branch.provision(user=self.job.user)


class SyncBranchJob(JobRunner):
    """
    Sync changes from main into a Branch.
    """
    class Meta:
        name = 'Sync branch'

    def _disconnect_signal_receivers(self):
        """
        Disconnect object change handlers before syncing.
        """
        post_save.disconnect(handle_changed_object)
        m2m_changed.disconnect(handle_changed_object)
        pre_delete.disconnect(handle_deleted_object)

    def _reconnect_signal_receivers(self):
        """
        Reconnect object change handlers after syncing.
        """
        post_save.connect(handle_changed_object)
        m2m_changed.connect(handle_changed_object)
        pre_delete.connect(handle_deleted_object)

    def run(self, commit=True, *args, **kwargs):
        # Initialize logging
        logger = logging.getLogger('netbox_branching.branch.sync')
        logger.setLevel(logging.DEBUG)
        logger.addHandler(ListHandler(queue=get_job_log(self.job)))

        # Disconnect changelog handlers
        self._disconnect_signal_receivers()

        # Sync the branch
        try:
            branch = self.job.object
            branch.sync(user=self.job.user, commit=commit)
        except AbortTransaction:
            logger.info("Dry run completed; rolling back changes")
        except Exception as e:
            # TODO: Can JobRunner be extended to handle this more cleanly?
            # Ensure that signal handlers are reconnected
            self._reconnect_signal_receivers()
            raise e

        # Reconnect signal handlers
        self._reconnect_signal_receivers()


class MergeBranchJob(JobRunner):
    """
    Merge changes from a Branch into main.
    """
    class Meta:
        name = 'Merge branch'

    def run(self, commit=True, *args, **kwargs):
        # Initialize logging
        logger = logging.getLogger('netbox_branching.branch.merge')
        logger.setLevel(logging.DEBUG)
        logger.addHandler(ListHandler(queue=get_job_log(self.job)))

        # Merge the Branch
        try:
            branch = self.job.object
            branch.merge(user=self.job.user, commit=commit)
        except AbortTransaction:
            logger.info("Dry run completed; rolling back changes")


class RevertBranchJob(JobRunner):
    """
    Revert changes from a merged Branch.
    """
    class Meta:
        name = 'Revert branch'

    def run(self, commit=True, *args, **kwargs):
        # Initialize logging
        logger = logging.getLogger('netbox_branching.branch.revert')
        logger.setLevel(logging.DEBUG)
        logger.addHandler(ListHandler(queue=get_job_log(self.job)))

        # Merge the Branch
        try:
            branch = self.job.object
            branch.revert(user=self.job.user, commit=commit)
        except AbortTransaction:
            logger.info("Dry run completed; rolling back changes")



# Name of the main (non-branch) PostgreSQL schema
MAIN_SCHEMA = 'public'

# HTTP cookie
COOKIE_NAME = 'active_branch'

# HTTP header for API requests
BRANCH_HEADER = 'X-NetBox-Branch'

# URL query parameter name
QUERY_PARAM = '_branch'
